/**
 *  Copyright notice
 *  
 *  This file is part of the Processing sketch `CavityCalculator' 
 *  http://www.gwoptics.org/processing/cavity_calculator/
 *
 *  Copyright (C) 2010 onwards Jamie Dougherty and Andreas Freise
 *  
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
 *  MA  02110-1301, USA.
 */
 
class Axis
  {
    int x1, x2;  //endpoints of axes in pixels
    int y1, y2;
    int axiszero;
    String axislabel;
    int textx, texty;
    int majorSpacing;
    int minorSpacing;
    AxisConfig axisConfig;
    float startval, incr_val;
    color axiscolor, textcolor, tickcolor, minorTickColor;
    int axisweight;
    int majorTickSize;
    int tickweight;
    int minor_weight;
    int minorTickSize;
    int fontsize;
    PFont font;
    int yzero,xzero;
    String formatchoice;
    Boolean isZeroTickDrawn;
    
    //CONSTRUCTOR:  When creating a new axis object, the starting and ending points of 
    //the line must be defined, just as when drawing a regular line.  Then the axis
    //label, the AxisConfig object being used, and the pixel coordinate which corresponds
    //to where the axis equals 0 (horizontal pixel for a horizontal axis, vertical pixel
    //for a vertical axis).
    Axis(int t_x1, int t_y1, int t_x2, int t_y2, String t_label, AxisConfig temp_AConfig, int t_zero)
    {
      x1 = t_x1;
      x2 = t_x2;
      y1 = t_y1;
      y2 = t_y2;
      majorSpacing = 20;
      minorSpacing = majorSpacing;                //default is to hide the minor ticks behind the major
      axislabel = t_label;
      axisConfig = temp_AConfig;
      axiszero = t_zero;
      axiscolor = axisConfig.axiscolor;
      textcolor = axisConfig.textcolor;
      tickcolor = axisConfig.tickcolor;
      minorTickColor = axisConfig.minorTickColor;
      axisweight = axisConfig.axisweight;
      majorTickSize = axisConfig.majorTickSize;
      textx = max(x1, x2)-30;                     //should put the text roughly in a good position
      texty = min(y1, y2)+20;                     //relative to the axes. Can be changed with setTextPosition
      tickweight = axisConfig.tickweight;
      minor_weight = axisConfig.minor_weight;
      minorTickSize = axisConfig.minorTickSize;
      fontsize = axisConfig.fontsize;
      font = axisConfig.font;
      startval = 0;                              //Default: Tick marks calculated starting from 0
      incr_val = 1;                              //Default: Tick mark labels increment by 1
      yzero = 0;                                 //Default: Zeroes are where Processing wants them,
      xzero = 0;                                 //the top left corner of the window.
      formatchoice = "%.0f";                     //No decimal places, truncates any numbers after decimal
      isZeroTickDrawn = false;
    }
  
    //This functions requires parameters of an x and a y coordinate to change the
    //location of the axis label from the default.
    //NOTE:  The y axis label is rotated 90 degrees.  This means the whole window coordinate
    //system, for just the y axis label, is rotated 90 degrees clockwise around the origin (0,0)
    //(the upper left corner of the screen).  So your x position will have to be negative here
    //and the y will have to be what your x in the normal coordinate system would be.
    void setTextPosition(int t_textx, int t_texty)
    {
      textx = t_textx;
      texty = t_texty;
      if((textx>width)||(textx<0)||(texty>height)||(texty<0))
      {
        println("Text placed outside the window range. Check your coordinates for text position.");
      }
    }
  
    //This function requires an integer number of pixels to be used as the spacing between 
    //the major tick marks (default spacing is 20 pixels).
    void setMajorSpacing(int temp_spacing)
    {
      majorSpacing = temp_spacing;
    }
  
    //This function requires an integer number of pixels to be used as the spacing between
    //the minor tick marks (default spacing is the same as the default major tick spacing, so
    //the minor ticks should be hidden if no spacings are defined).  
    void setMinorSpacing(int t_minorSpacing)
    {
      minorSpacing = t_minorSpacing;
      //If minorSpacing > majorSpacing, could lead to problems with tick marks, so should let the user know.
      if(minorSpacing>majorSpacing) {println("minorSpacing > majorSpacing. Check your spacing values, or just don't draw minor ticks if not wanted."); }
    }
  
    //This function requires two float values. The first is the value from which the tick mark labels will 
    //start. If there is an origin on the screen, this should be 0. If the origin is not shown on the graph,
    //the starting value should be the leftmost(smallest) value. The second parameter is the value by which
    //the tick mark labels are to increment.
    void setMajorLabels(float t_startval, float t_incr_val)
    {
      startval = (t_startval);
      incr_val = (t_incr_val);
    }
    
    //This function allows the tick mark labels to be formatted as integers, floats, or scientific notation.
    //For example, the format string "%.2f" will be in the format of a float truncated after two places after
    //the decimal. "%.1e" will output a number such as 1.0e3, whereas "%.1E" will output 1.0E3. The default is
    //an integer in regular notation (".0f").
    //More formatting options can be found online.
    void setTickLblFormat(String temp_format)
    {
      formatchoice = temp_format;
      if(formatchoice==""){ throw new RuntimeException("The format choice cannot be empty");}
    }
    
    void isZeroTickDrawn(Boolean t_iszerodrawn)
    {
      isZeroTickDrawn = t_iszerodrawn;
    }
  
    //This function simply draws the line of the axis and the label, no parameters
    void drawAxis()
    {
      stroke(axiscolor);
      strokeWeight(axisweight);
      line(x1,y1,x2,y2);
      textFont(font,fontsize);
      fill(textcolor);
      //text(axislabel, textx, texty);
      if((y1-y2)==0)
      {
        text(axislabel, textx, texty);
      }
      if((x1-x2)==0)            //rotate the y label
      {
        rotate(-PI/2);
        text(axislabel, -texty, textx);    //this compensates for the rotation, but keeps the intuition the same for the user
        rotate(PI/2);
      }
    }
  
    //Tick marks are drawn from a starting point at wherever the user defines
    //the zero point of the axis to be. It draws ticks until it hits the ends of
    //the axis
  
    //This function draws the major tick marks, no parameters.  
    //If you don't want major tick marks, simply don't call this function.
    void drawMajorTicks()
    {
      //distinguish between x and y axes
      if((y1-y2)==0)                //if it's the x axis (horizontal), do...
      {
        yzero = axiszero;
        
        if(isZeroTickDrawn)
        {
          stroke(tickcolor);
          strokeWeight(tickweight);
          line(axiszero,((abs(y2+y1)/2)-(majorTickSize/2)),axiszero,((abs(y2+y1)/2)+(majorTickSize/2)));
        }
        
        for(int i=1; (axiszero+i*majorSpacing)<=max(x1,x2); i++)      //major ticks going right...
        {
          stroke(tickcolor);
          strokeWeight(tickweight);
          line(axiszero+(i*majorSpacing),((abs(y2+y1)/2)-(majorTickSize/2)),axiszero+(i*majorSpacing),((abs(y2+y1)/2)+(majorTickSize/2)));
        }
      
        for(int i=1; (axiszero-i*majorSpacing)>=min(x1,x2); i++)      //going left...
        {
          stroke(tickcolor);
          strokeWeight(tickweight);
          line(axiszero-(i*majorSpacing),((abs(y2+y1)/2)-(majorTickSize/2)),axiszero-(i*majorSpacing),((abs(y2+y1)/2)+(majorTickSize/2)));
        }
      }
      
      if((x2-x1)==0)                //if it's the y (vertical) axis, do...
      {
        yzero = axiszero;
        
        if(isZeroTickDrawn)
        {
          stroke(tickcolor);
          strokeWeight(tickweight);
          line(((abs(x2+x1)/2)-(majorTickSize/2)),axiszero,((abs(x2+x1)/2)+(majorTickSize/2)),axiszero);
        }
        
        for(int i=1; (axiszero+(i*majorSpacing))<=max(y1,y2); i++)    //going down...
        {
          stroke(tickcolor);
          strokeWeight(tickweight);
          line(((abs(x2+x1)/2)-(majorTickSize/2)),axiszero+(i*majorSpacing),((abs(x2+x1)/2)+(majorTickSize/2)),axiszero+(i*majorSpacing));
        }
    
        for(int i=1; (axiszero-(i*majorSpacing))>=min(y1,y2); i++)    //going up...
        {
          stroke(tickcolor);
          strokeWeight(tickweight);
          line(((abs(x2+x1)/2)-(majorTickSize/2)),axiszero-(i*majorSpacing),((abs(x2+x1)/2)+(majorTickSize/2)),axiszero-(i*majorSpacing));
        }
      }
      
      if(((y1-y2)!=0)&&((x1-x2)!=0))
      {
        throw new RuntimeException("Each axis must be perfectly horizontal or vertical. Check the coordinates passed to the Axis objects.");
      }
    }
  
    //This function draws the minor tick marks, no parameters.  
    //If you don't want minor tick marks, simply don't call this function.  
    void drawMinorTicks()
    {
      //distinguish between x and y axes
      if((y2-y1)==0)                //if it's the x axis (horizontal), do...
      {
        for(int i=1; (axiszero+i*minorSpacing)<=max(x1,x2); i++)      //minor ticks going right...
        {
          stroke(minorTickColor);
          strokeWeight(tickweight);
          line(axiszero+(i*minorSpacing),((abs(y2+y1)/2)-(minorTickSize/2)),axiszero+(i*minorSpacing),((abs(y2+y1)/2)+(minorTickSize/2)));
        }
      
        for(int i=1; (axiszero-i*minorSpacing)>=min(x1,x2); i++)      //going left...
        {
          stroke(minorTickColor);
          strokeWeight(tickweight);
          line(axiszero-(i*minorSpacing),((abs(y2+y1)/2)-(minorTickSize/2)),axiszero-(i*minorSpacing),((abs(y2+y1)/2)+(minorTickSize/2)));
        }
      }
    
      if((x2-x1)==0)                //if it's the y (vertical) axis, do...
      {
        for(int i=1; (axiszero+(i*minorSpacing))<=max(y1,y2); i++)    //going down...
        {
          stroke(minorTickColor);
          strokeWeight(tickweight);
          line(((abs(x2+x1)/2)-(minorTickSize/2)),axiszero+(i*minorSpacing),((abs(x2+x1)/2)+(minorTickSize/2)),axiszero+(i*minorSpacing));
        }
    
        for(int i=1; (axiszero-(i*minorSpacing))>=min(y1,y2); i++)    //going up...
        {
          stroke(minorTickColor);
          strokeWeight(tickweight);
          line(((abs(x2+x1)/2)-(minorTickSize/2)),axiszero-(i*minorSpacing),((abs(x2+x1)/2)+(minorTickSize/2)),axiszero-(i*minorSpacing));
        }
      }
    }
  
    //If this function is called, it draws the tick mark labels as defined by the user for the starting value (at axiszero)
    //and the increment value (defined via setMajorLabels()). Not perfect... should fix things about font size for the labels,
    //since it's currently the same size as the axis labels.  
    void drawMajorTickLabels()
    {
      //distinguish between x and y axes
      if((y2-y1)==0)                //if it's the x axis (horizontal), do...
      {
        if(isZeroTickDrawn)
        {
          textFont(font,fontsize);
          fill(textcolor);
          String zerotick = "0";
          zerotick = String.format(formatchoice,Float.valueOf(zerotick));
          text(zerotick,axiszero,(abs(y2+y1)/2)+(majorTickSize/2)+textAscent());
        }
        
        for(int i=1; (axiszero+i*majorSpacing)<=max(x1,x2); i++)
        {
          String tickLbl = String.valueOf(startval+incr_val*i);
          textFont(font,fontsize);
          fill(textcolor);
          String fullTickLbl = String.format(formatchoice,Float.valueOf(tickLbl));
          //text(fullTickLbl, axiszero+(i*majorSpacing)-textWidth(fullTickLbl)/2, (abs(y2+y1)/2)+(majorTickSize/2)+textAscent());  //**********************
          text(fullTickLbl, axiszero+(i*majorSpacing), (abs(y2+y1)/2)+(majorTickSize/2)+textAscent());
        }
      
        for(int i=1; (axiszero-i*majorSpacing)>=min(x1,x2); i++)      //going left...
        {
          String tickLbl = String.valueOf(startval-incr_val*i);
          textFont(font,fontsize);
          fill(textcolor);
          String fullTickLbl = String.format(formatchoice,Float.valueOf(tickLbl));
          //text(fullTickLbl,axiszero-(i*majorSpacing)-textWidth(fullTickLbl)/2, (abs(y2+y1)/2)+(majorTickSize/2)+textAscent());    //**********************
          text(fullTickLbl,axiszero-(i*majorSpacing), (abs(y2+y1)/2)+(majorTickSize/2)+textAscent());
        }
      }
    
      if((x2-x1)==0)                //if it's the y (vertical) axis, do...
      {
        if(isZeroTickDrawn)
        {
          textFont(font,fontsize);
          fill(textcolor);
          String zerotick = "0";
          zerotick = String.format(formatchoice,Float.valueOf(zerotick));
          text(zerotick,(abs(x2+x1)/2)-majorTickSize/2-textWidth(zerotick),axiszero);
        }
        
        for(int i=1; (axiszero+(i*majorSpacing))<=max(y1,y2); i++)    //going down...
        {
          String tickLbl = String.valueOf(startval-incr_val*i);
          textFont(font,fontsize);
          fill(textcolor);
          String fullTickLbl = String.format(formatchoice,Float.valueOf(tickLbl));
          text(fullTickLbl, (abs(x2+x1)/2)-majorTickSize/2-textWidth(fullTickLbl), axiszero+(i*majorSpacing)+(fontsize/4));
        }
    
        for(int i=1; (axiszero-(i*majorSpacing))>=min(y1,y2); i++)    //going up...
        {
          String tickLbl = String.valueOf(startval+incr_val*i);
          textFont(font,fontsize);
          fill(textcolor);
          String fullTickLbl = String.format(formatchoice,Float.valueOf(tickLbl));
          text(fullTickLbl, (abs(x2+x1)/2)-majorTickSize/2-textWidth(fullTickLbl), axiszero-(i*majorSpacing)+(fontsize/4));
        }
      }
    }
    
    ///Get functions. Self-explanatory.
    int getXmax()
    {
      return max(x1,x2);
    }
    
    int getXmin()
    {
      return min(x1,x2);
    }
    
    int getYmin() { return min(y1,y2); }
    int getYmax() { return max(y1,y2); }
    
    int getXzero() { return xzero; }
    int getYzero() { return yzero; }
  }
