/**
 *  Copyright notice
 *  
 *  This file is part of the Processing sketch `CavityCalculator' 
 *  http://www.gwoptics.org/processing/cavity_calculator/
 *
 *  Copyright (C) 2010 onwards Jamie Dougherty and Andreas Freise
 *  
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
 *  MA  02110-1301, USA.
 */

import org.gwoptics.ValueType;
import guicomponents.*;
import org.gwoptics.Logo;
import org.gwoptics.LogoSize;
import guicomponents.GCheckbox;

//declare variables, objects, and fonts
GWSlider R1slider;
GTextField R1textfield;

GWSlider R2slider;
GTextField R2textfield;

GWSlider RoCslider;
GTextField RoCtextfield;

GWSlider delta_x_slider;

LEDMeter meter;
LEDMeterAxis meterAxis;

AxisConfig axisconfig;
PlotConfig plotconfig;
TraceConfig traceConfig1, traceConfig2, traceConfig3;
PlotAreaBackground area1;
Axis xaxis,yaxis;
Plotting alltraces;
InternalPower00 mode00;
InternalPower10 mode10;
InternalPower20 mode20;
PFont font;
float R1,R2,delta_x,L,RoC;
PImage cavityimage;
Logo _logo;
PFont titleFont;
PFont expFont;
PFont meterTitleFont;
String explanationtext;
GCheckbox chkWhiteLine;

void setup()
{
  size(900,600);
  smooth();
  L = 1;        //length of the cavity is fixed to 1 meter
  
  //initialize the axis, plot, and trace config objects
  axisconfig = new AxisConfig();
  plotconfig = new PlotConfig();
  traceConfig1 = new TraceConfig();
  traceConfig2 = new TraceConfig();
  traceConfig3 = new TraceConfig();
  
  //set the config settings
  plotconfig.bordercolor = color(255);
  plotconfig.fillcolor = color(0);
  plotconfig.borderweight = 3;
  axisconfig.axiscolor = color(255);
  axisconfig.textcolor = color(255);
  axisconfig.tickcolor = color(255);
  traceConfig1.tracecolor = color(255,0,0);    //each trace has its own color so its own config object
  traceConfig1.traceweight = 2;
  traceConfig2.tracecolor = color(0,255,0);
  traceConfig2.traceweight = 2;
  traceConfig3.tracecolor = color(0,0,255);
  traceConfig3.traceweight = 2;
  
  //Initialize the Axes and PlotAreaBackground
  area1 = new PlotAreaBackground(5,15,540,389,plotconfig.getConfig());    //x pos, y pos, width, height, config object
  xaxis = new Axis(60,350,526,350,"delta x (degrees)",axisconfig.getConfig(),293);  //x1,y1,x2,y2,axis label, config, axis zero
  yaxis = new Axis(60,25,60,350,"log of Intensity",axisconfig.getConfig(),188);
  
  //Tick label formatting
  xaxis.setTickLblFormat("%.0f");    //no decimal places for the ticks
  xaxis.setMajorSpacing(33);         //(length of x axis in pixels)/(its length in graph space) = 33.28571429px/graph unit
  yaxis.setMajorSpacing(54);
  xaxis.setMajorLabels(0,90);        //major tick labels start at 0 and increment by 90
  yaxis.setMajorLabels(0,1);         //major tick labels start at 0 and increment by 1
  xaxis.isZeroTickDrawn(true);
  yaxis.isZeroTickDrawn(true);
  
  //Axis label text formatting
  xaxis.setTextPosition(300,400);
  yaxis.setTextPosition(25,200);
  
  //Initialize the trace objects
  alltraces = new Plotting(xaxis,yaxis);    //need to pass the alltraces object the axes that you'll be drawing the traces on
  mode00 = new InternalPower00();           //these are the internal power equations being used, one for each mode.
  mode10 = new InternalPower10();
  mode20 = new InternalPower20();
  
  alltraces.addTrace(mode00,traceConfig1);  //tell alltraces which traces and configs you want
  alltraces.addTrace(mode10,traceConfig2);
  alltraces.addTrace(mode20,traceConfig3);
  
  //Initialize and add all the sliders
  R1slider = new GWSlider(this,200,559,110);
  R1slider.setTickColour(255,255,255);
  R1slider.setFontColour(255,255,255);
  R1slider.setValueType(GWSlider.DECIMAL);
  R1slider.setLimits(0.96,0.00,0.99);        //1.00 reflectivity is unrealistic, so we set the limit to 0.99
  R1slider.setTickCount(5);
  R1slider.setPrecision(2);
  R1textfield = new GTextField(this,String.valueOf(0.96),160,557,36,15);  //This sets the textfield value for before you move the slider.
  
  R2slider = new GWSlider(this,370,559,110);
  R2slider.setTickColour(255,255,255);
  R2slider.setFontColour(255,255,255);
  R2slider.setValueType(GWSlider.DECIMAL);
  R2slider.setLimits(0.96,0,0.99);
  R2slider.setTickCount(5);
  R2slider.setPrecision(2);
  R2textfield = new GTextField(this,String.valueOf(0.96f),332,557,36,15);
  
  RoCslider = new GWSlider(this,370,452,110);
  RoCslider.setTickColour(255,255,255);
  RoCslider.setFontColour(255,255,255);
  RoCslider.setValueType(GWSlider.DECIMAL);
  RoCslider.setLimits(0.5,0,1);                  //represents 1/RoC, so the RoC can range from L to infinity
  RoCslider.setTickCount(5);
  RoCslider.setPrecision(2);
  RoCtextfield = new GTextField(this,String.valueOf(0.50f),332,450,36,15);
  
  delta_x_slider = new GWSlider(this,54,368,478);  //doesn't exactly line up with axis but close
  delta_x_slider.setTickColour(255,255,255);
  delta_x_slider.setLimits(0,-3.5*PI,3.5*PI);      //default position for slider, min value, max value
  delta_x_slider.setTickCount(13);
  delta_x_slider.setValueType(GWSlider.DECIMAL);
  delta_x_slider.setPrecision(2);
  delta_x_slider.setRenderMaxMinLabel(false);
  delta_x_slider.setRenderValueLabel(false);
  
  meter = new LEDMeter(557,490,349,30,20,0.0,1.0);  //x,y,length,num,height,min,max
  meter.setGradient();
  meter.setGapSize(0);
  meterAxis = new LEDMeterAxis(557,500,10,33,0,1);  //xpos, ypos, num labels, xstep, min, max
  meterAxis.setFontSize(12);
  
  //Importing images
  cavityimage = loadImage("cavitypng.png");
  _logo = new Logo(this, 770,570,true,LogoSize.Size25);
  
  //Create fonts
  titleFont = createFont("verdana",20);
  meterTitleFont = createFont("verdana",13);
  expFont = createFont("verdana",11);
  
  //Write the explanation here
  explanationtext = "This applet graphs the internal power of an optical cavity as a function of tuning for transverse modes 00 (red), 10 (green), and 20 (blue)."+
  "The slider below the graph controls the tuning of the second mirror. Below, "+
  "the meter shows the percentage of the input power that is outputted through the second mirror at the current value of delta x (the tuning). "+
  "The cavity is drawn below the graph. Each mirror has a reflectivity slider. The 1/RoC slider controls the inverse "+
  "of the Radius of Curvature of the second mirror. 1/RoC=0 represents a flat mirror (infinite RoC), and 1/RoC=1=L represents a radius of "+
  "curvature equal to the length of the cavity (1m). "+
  "Using the sliders, you should be able to notice changes in the Guoy phase and the FSR for the different modes. The sliders can be controlled with the mouse or "+
  "left/right arrow keys.";
  
  //Adds a checkbox to select whether or not we want the white line on the delta_x slider to show. Defaulted to yes/true
  chkWhiteLine = new GCheckbox(this,"Show white line for delta x",550,380,200);
  chkWhiteLine.setSelected(true);
}

void draw()
{
  background(200,200,200);
  
  //Draw plotting area and the traces
  area1.drawRect();
  xaxis.drawAxis();
  yaxis.drawAxis();
  xaxis.drawMajorTicks();
  yaxis.drawMajorTicks();
  xaxis.drawMajorTickLabels();
  yaxis.drawMajorTickLabels();
  alltraces.plotTraces(-3.5*PI,3.5*PI,-3,3);    //need to know the x and y axis ranges in graph space
  
  //Draw the cavity image
  image(cavityimage,10,412);
  
  //Assign the slider values to variables
  R1 = R1slider.getValuef();
  R2 = R2slider.getValuef();
  delta_x = delta_x_slider.getValuef();
  RoC = 1/(RoCslider.getValuef());
  
  //Set and write the Title to the screen
  textFont(titleFont);
  fill(0);
  text("Optical Resonator Calculator",720,30);
  
  //Set and write a label for the meter
  textFont(meterTitleFont);
  fill(0);
  text("Percent of Input Power Transmitted",720,480);
  
  //Explanation box
  stroke(255);
  fill(190,204,223);
  rect(575,45,300,328);
  textFont(expFont);
  fill(0);
  text(explanationtext,579,49,295,340);
  
  //Add logo
  _logo.draw();
  
  //Power meter.  Apparently in Processing log is natural log... Also now each beam has an input power of 1/3 W, so the total input power is 1W.
  float sum_delta_x = (pow(10,mode00.computePoint(delta_x))+pow(10,mode10.computePoint(delta_x))+pow(10,mode20.computePoint(delta_x)))/3;
  float transmitted_power = sum_delta_x*(1-R2);
  meter.setValue(transmitted_power);
  meter.response();
  meterAxis.display();

  //Adjusting the line drawn off the slider for rounding errors (there's most definitely a more elegant way to 
  //do this, but this works well enough)
  if(chkWhiteLine.isSelected()){
 /*
    if(delta_x_slider.getValuef() < -5.6)
    { line((delta_x_slider.getValuef()*float(480/22))+293,368,(delta_x_slider.getValuef()*float(482/22))+293,15);}
    if((delta_x_slider.getValuef() >= -5.6)&&(delta_x_slider.getValuef() <= 0))
    { line((delta_x_slider.getValuef()*float(480/22))+294,368,(delta_x_slider.getValuef()*float(482/22))+294,15);}
    if((delta_x_slider.getValuef() > 0)&&(delta_x_slider.getValuef() < 5.5))
    { line((delta_x_slider.getValuef()*float(480/22))+295,368,(delta_x_slider.getValuef()*float(482/22))+295,15);}
    if((delta_x_slider.getValuef() >= 5.5))
    */
    line((delta_x_slider.getValuef()*float(478/22))+293,368,(delta_x_slider.getValuef()*float(478/22))+293,15);
  }
}

//These are the functions that are being plotted. All the same except for different values of n
class InternalPower00 implements GraphCallback
{
  int n,m;
  float Power00;
  float zR, psiRT;
  public float computePoint(float t_delta_x)
  {
    delta_x = t_delta_x;
    n = 0;
    m = 0;
    zR = sqrt(L*(RoC-L));
    psiRT = 2*atan(L/zR);
    //Need to convert the natural log to log base 10...
    Power00 = log((1-R1)/(1+(R1*R2)-(2*sqrt(R1)*sqrt(R2)*cos(delta_x+(psiRT*(n+m+1))))))/log(10);
    return Power00;
  }
}
  
class InternalPower10 implements GraphCallback
{
  int n,m;
  float Power10;
  float zR, psiRT;
  public float computePoint(float t_delta_x)
  {
    delta_x = t_delta_x;
    n = 1;
    m = 0;
    zR = sqrt(L*(RoC-L));
    psiRT = 2*atan(L/zR);
    Power10 = (log((1-R1)/(1+(R1*R2)-(2*sqrt(R1)*sqrt(R2)*cos(delta_x+(psiRT*(n+m+1)))))))/log(10);
    return Power10;
  }
}
  
class InternalPower20 implements GraphCallback
{
  int n,m;
  float Power20;
  float zR, psiRT;
  public float computePoint(float t_delta_x)
  {
    delta_x = t_delta_x;
    n = 2;
    m = 0;
    zR = sqrt(L*(RoC-L));
    psiRT = 2*atan(L/zR);
    Power20 = (log((1-R1)/(1+(R1*R2)-(2*sqrt(R1)*sqrt(R2)*cos(delta_x+(psiRT*(n+m+1)))))))/log(10);
    return Power20;
  }
}

//This method tells the prorgam what to do if the user moves the sliders
void handleSliderEvents(GSlider slider){
  float precision = 100;
  if(slider == R1slider){
    R1textfield.setText(String.format("%.2f",slider.getValuef()));
  }
  if(slider == R2slider){
    R2textfield.setText(String.format("%.2f",slider.getValuef()));  
  }
  if(slider == RoCslider){
    RoCtextfield.setText(String.format("%.2f",slider.getValuef()));
  }
}

//This method tells the program what to do if the user types a value into a textfield and hits 'Enter'
void handleTextFieldEvents(GTextField tfield){
  String line1="";
  float temp_val;
  if(tfield == R1textfield){
    switch (R1textfield.eventType){
    case GTextField.CHANGED:
      break;
    case GTextField.SET:
      break;
    case GTextField.ENTERED:
      line1=R1textfield.getText();
      temp_val = parseFloat(line1);
      R1slider.setValue(temp_val);
      break;
    }
  }
  if(tfield == R2textfield){
    switch (R2textfield.eventType){
    case GTextField.CHANGED:
      break;
    case GTextField.SET:
      break;
    case GTextField.ENTERED:
      line1=R2textfield.getText();
      temp_val =  parseFloat(line1);
      R2slider.setValue(temp_val);
      break;
    }
  }
  if(tfield == RoCtextfield){
    switch (RoCtextfield.eventType){
    case GTextField.CHANGED:
      break;
    case GTextField.SET:
      break;
    case GTextField.ENTERED:
      line1=RoCtextfield.getText();
      temp_val =  parseFloat(line1);
      RoCslider.setValue(temp_val);  
      break;
    }
  }
}
