/**
 *  Copyright notice
 *  
 *  This file is part of the Processing sketch `CavityCalculator' 
 *  http://www.gwoptics.org/processing/cavity_calculator/
 *
 *  Copyright (C) 2010 onwards Jamie Dougherty and Andreas Freise
 *  
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
 *  MA  02110-1301, USA.
 */

class LEDMeter {
  
  // Primary variables
  int xp, yp, s_length, height_led, num_leds;
  float min_value, max_value;
  
  // Secondary variables
  int leds_lit, colour1_lim, colour2_lim, colour3_lim;
  color lit_colour1, dark_colour1, lit_colour2, dark_colour2;
  color lit_colour3, dark_colour3;
  int gap_size;
  float increment;
  int grad;
  
  // Images and objects
  PImage unlit_led_image, lit_led_image;
  LEDMeterAxis axis;

  
  // Other variables
  int width_led;
  
  // Value represented by the meter
  float meter_value;
 
  
  LEDMeter(int t_xp, int t_yp, int t_s_length, int t_num_leds, 
            int t_height_led, float t_min_value, float t_max_value) {
              
              xp = t_xp;
              yp = t_yp;
              s_length = t_s_length;
              num_leds = t_num_leds;
              height_led = t_height_led;
              min_value = t_min_value;
              max_value = t_max_value;
                 
              // Default values
              gap_size = 5;
              // Initial number of leds lit up
              leds_lit = round(num_leds/2.0);
              colourLimits();
              
              // Default colours
              defaultColours();
              
              unlit_led_image = loadImage("led_unlit.png");
              lit_led_image = loadImage("led_lit2.png");


              
              width_led = (int)((s_length-(gap_size*(num_leds-1)))/(float)num_leds);
              
              axis = new LEDMeterAxis(xp,yp+height_led,
                                    num_leds, width_led+gap_size,min_value,max_value);
              
              increment = (max_value-min_value)/(float)num_leds;
              
              meter_value = leds_lit*increment+min_value;
              
              grad = 0;
              
            }
  
  // Set functions for secondary variables
  
  void setColours(color t_colour1, color t_colour2,
                         color t_colour3) {
    lit_colour1 = t_colour1;
    lit_colour2 = t_colour2;
    lit_colour3 = t_colour3;
    dark_colour1 = darkColours(lit_colour1);
    dark_colour2 = darkColours(lit_colour2);
    dark_colour3 = darkColours(lit_colour3);
  } 
  
  void setGapSize(int t_gap_size) {
    gap_size = t_gap_size;
    width_led = (int)((s_length-(gap_size*(num_leds-1)))/(float)num_leds);
  }
  
  void setValue(float value) {
    if (value>max_value)
      meter_value = max_value;
    else if (value<min_value)
      meter_value = min_value;
    else
      meter_value = value;
    leds_lit = round((meter_value-min_value)/increment);
  }
  
  void setLedsLit(int t_leds_lit) {
    if (t_leds_lit > num_leds)
      leds_lit = num_leds;
    else if (t_leds_lit <0)
      leds_lit = 0;
    else 
      leds_lit = t_leds_lit;
    meter_value = leds_lit*increment+min_value;
  }
  
  void setGradient() {
    grad = 1;
    colour1_lim = floor(colour1_lim+((colour2_lim-colour1_lim)/2.0));
    colour2_lim = num_leds-1;
  }
  
  void setColourLimits(int t_colour1_lim, int t_colour2_lim, int t_colour3_lim) {
    colour1_lim = t_colour1_lim;
    colour2_lim = t_colour2_lim;
    colour3_lim = t_colour3_lim;
  }
  
  void setAxisTextColour(color axis_text_colour) {
    axis.setTextColour(axis_text_colour);
  }
  
  void setAxisFontSize(int axis_font_size) {
    axis.setFontSize(axis_font_size);
  }
    
  void colourLimits() {
    int colour1_leds, colour2_leds, colour3_leds;
    int leds_per_colour = round(num_leds/3.0);
    
    if (leds_per_colour >= num_leds/3.0) {
      colour1_leds = leds_per_colour;
      colour2_leds = leds_per_colour;
      if (leds_per_colour == num_leds/3.0)
        colour3_leds = leds_per_colour;
      else 
        colour3_leds = leds_per_colour - 1;
      }
    
    else {
      colour1_leds = leds_per_colour+1;
      colour2_leds = leds_per_colour;
      colour3_leds = leds_per_colour;
    }
    
    colour1_lim = colour1_leds;
    colour2_lim = colour1_leds+colour2_leds;
    colour3_lim = colour1_leds+colour2_leds+colour3_leds;  
  }
  
  void defaultColours() {
    lit_colour1 = color(0,255,0);
    lit_colour2 = color(255,255,0);
    lit_colour3 = color(255,0,0);
    dark_colour1 = darkColours(lit_colour1);
    dark_colour2 = darkColours(lit_colour2);
    dark_colour3 = darkColours(lit_colour3);
  }
  
  color darkColours(color lit_colour) {
    float red_f = red(lit_colour);
    float green_f = green(lit_colour);
    float blue_f = blue(lit_colour);
    
    float dark_red_f = dark(red_f);
    float dark_green_f = dark(green_f);
    float dark_blue_f = dark(blue_f);
    
    color dark_colour = color(dark_red_f,dark_green_f,dark_blue_f);
    return dark_colour;
  }
  
  float dark(float seperate_colour) {
    float dark_colour;
    if (seperate_colour == 0) 
      dark_colour = 30;
    else 
      dark_colour = 0.45*seperate_colour;
    return dark_colour; 
  }
  
  
  void displayAxis() {
    axis.display();
  }
  
  void response() {
    sliderResponse();
    drawLED();
  }
  
  void sliderResponse() {
    float gradient_fraction1 = 1/(float)colour1_lim;
    float gradient_fraction2 = 1/(float)(colour2_lim-colour1_lim);

    for (int i=0; i<num_leds; i++) { 
      
      color lit_grad_colour1 = mergeColour(lit_colour1,lit_colour2,gradient_fraction1*i*grad);
      color lit_grad_colour2 = mergeColour(lit_colour2,lit_colour3,gradient_fraction2*(i-colour1_lim+1)*grad);
      
      color dark_grad_colour1 = darkColours(lit_grad_colour1);
      color dark_grad_colour2 = darkColours(lit_grad_colour2);
      
      if (i < leds_lit) {
        if (i < colour1_lim)
          fill(lit_grad_colour1);
        else if (i < colour2_lim)
          fill(lit_grad_colour2);
        else 
          fill(lit_colour3);
      }
      else
        if (i < colour1_lim)
          fill(dark_grad_colour1);
        else if (i < colour2_lim)
          fill(dark_grad_colour2);
        else 
          fill(dark_colour3);
      
      rect(xp+((width_led+gap_size)*i), yp, width_led, height_led);
    }
  }
  
  color mergeColour(color colour1, color colour2, float merge_fraction) {
    float red1 = red(colour1);
    float green1 = green(colour1);
    float blue1 = blue(colour1);
    float red2 = red(colour2);
    float green2 = green(colour2);
    float blue2 = blue(colour2);
    
    float merge_red = red1*(1-merge_fraction)+red2*merge_fraction;
    float merge_green = green1*(1-merge_fraction)+green2*merge_fraction;
    float merge_blue = blue1*(1-merge_fraction)+blue2*merge_fraction;
    
    color merge_colour = color(merge_red,merge_green,merge_blue);
    return merge_colour;
  }
  
  // Function to overlay images over rectangles
  void drawLED() {
    pushMatrix(); 
    float i_width = unlit_led_image.width;
    float i_height = unlit_led_image.height;
  
    translate(xp,yp);
    float w_scale = (width_led+1)/i_width;
    float h_scale = (height_led+1)/i_height;
    scale(w_scale,h_scale);
    //tint(255,200);
  
    for (int i=0; i<num_leds; i++) {
      if (i >= leds_lit)
        image(unlit_led_image, 0, 0);
      else
        image(lit_led_image, 0, 0);
      translate((width_led+gap_size)/w_scale,0);
    }
    popMatrix();
  }
  
  // Get functions
  
  int getGapSize() {
    return gap_size;
  }
  
  float getValue() {
    return meter_value;
  }
  
  int getLEDWidth() {
    return width_led;
  }
  
  

}
   
   
   
              
              
