/**
 *  Copyright notice
 *  
 *  This file is part of the Processing sketch `CavityCalculator' 
 *  http://www.gwoptics.org/processing/cavity_calculator/
 *  
 *  Copyright (C) 2010 onwards Jamie Dougherty and Andreas Freise
 *  
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
 *  MA  02110-1301, USA.
 */

class Plotting
{
  ArrayList traceArray;
  ArrayList traceconfigArray;
  TraceConfig traceConfig;

  private Axis _xAxis;    //added per Daniel's advice
  private Axis _yAxis;
 
  float yscale;
  float xscale;
  
  color traceColor;
  int traceWeight;
  
  //CONSTRUCTOR:  Requires an two Axis objects as its parameters (horizontal axis first,
  //then the vertical axis)
  Plotting(Axis X, Axis Y)    //have to define an x and y axis
  {
    traceArray = new ArrayList();    //creates an empty ArrayList, dynamic size
    traceconfigArray = new ArrayList();
    traceConfig = new TraceConfig();
    traceColor = traceConfig.tracecolor;
    traceWeight = traceConfig.traceweight;
    
    if(X == null)
      throw new RuntimeException("The X Axis object cannot be null");
    if(Y==null)
      throw new RuntimeException("The Y Axis object cannot be null");
      
    _xAxis = X;
    _yAxis = Y;
  }

  //Add a trace to a graphing region, requires an equation class that implements 
  //GraphCallback and have a computePoint(float x) function, and a traceconfig
  //object.
  public void addTrace(GraphCallback temp_cb, TraceConfig temp_traceconfig)
  {
    if(temp_cb==null) 
      throw new RuntimeException("The GraphCallback equation cannot be null");
    if(temp_traceconfig==null)
      throw new RuntimeException("The TraceConfig object cannot be null");
    traceconfigArray.add(temp_traceconfig);
    //Add temp_cb to an array to keep track of different traces
    traceArray.add(temp_cb);
  }

  //This function requires the minimum horizontal tick value, the maximum horizontal tick value, and
  //then the same for the y ticks.  It then handles the plotting of all the traces for that particular
  //graphing region.
  public void plotTraces(float t_minXtick, float t_maxXtick, float t_minYtick, float t_maxYtick)
  {
    int xmax=_xAxis.getXmax();    //This calls directly back to the xaxis object from the user code and calls its getXmax() function.
    int xmin=_xAxis.getXmin();
    int ymax=_yAxis.getYmax();
    int ymin=_yAxis.getYmin();
    int yzero=_yAxis.getYzero();
    int xzero=_xAxis.getXzero();
    float minXtick, minYtick, maxXtick, maxYtick;
    
    minXtick = t_minXtick;
    maxXtick = t_maxXtick;
    maxYtick = t_maxYtick;
    minYtick = t_minYtick;
    
    if(minXtick==maxXtick)
      throw new RuntimeException("Invalid x tick range. minXtick cannot = maxXtick. Check values passed to Plotting.plotTraces(float,float,float,float) function");
    if(minYtick==maxYtick)
      throw new RuntimeException("Invalid y tick range. minYtick cannot = maxYtick. Check values passed to Plotting.plotTraces(float,float,float,float) function");
    
    //not necessarily an error... they might want an inverted axis?
    if(minXtick>maxXtick)
      { println("The min x tick is larger than the max x tick. Graph will be inverted along the horizontal axis."); }
    if(minYtick>maxYtick)
      { println("The min y tick is larger than the max y tick. Graph will be inverted along the vertical axis."); }
    
    xscale = (xmax-xmin)/(maxXtick - minXtick);  //units: pixels/graph unit
    yscale = (ymax-ymin)/(maxYtick - minYtick);
    
    //will iterate along x for each trace in the trace array,
    //so two for loops
    for(int j=0; j < traceArray.size(); j++)
    {
      GraphCallback trace = (GraphCallback)traceArray.get(j);  //Have to cast the output
      TraceConfig traceconf = new TraceConfig();
      traceconf = (TraceConfig)traceconfigArray.get(j);
      
      //then a loop going over the axis
      //i increments in pixel space...
      //This has higher resolution than the one commented out below. Not as fast though...
      //For each pixel, it converts the pixel to graph space and calculates a tiny step in graph space around that point.
      //It then calculates a y value for each of these tiny points, and chooses the max one as the one to plot for the initial
      //x point.
      for(int i = xmin; i<xmax; i++)
      {
        float x = minXtick+(i-xmin)/xscale;      //converts i to its graph space equivalent, x
        float resolutionDivision = abs(x-(minXtick+(i+1-xmin)/xscale))/7;    //distance (in graph units) between checkpoints 
        float yplot = -trace.computePoint(x);
        float x2 = minXtick+(i+1-xmin)/xscale;
        float yplot2 = -trace.computePoint(x2);
        for(int k=-6; k<=6; k++)
        {
          float xcheckpoint = x+(k*resolutionDivision);
          float ycheckpoint = -trace.computePoint(xcheckpoint);      //still in graph space. - inverts it correctly
          if(ycheckpoint < yplot) { yplot = ycheckpoint; }
          
          float xcheckpoint2 = x2+(k*resolutionDivision);
          float ycheckpoint2 = -trace.computePoint(xcheckpoint2);
          if(ycheckpoint2 < yplot2) { yplot2 = ycheckpoint2; }
        }

        if(((yplot*yscale+yzero>=ymin)&&(yplot*yscale+yzero<=ymax))||((yplot2*yscale+yzero>=ymin)&&(yplot2*yscale+yzero<=ymax)))
        {
          TraceConfig traceconfig = (TraceConfig)traceconfigArray.get(j);
          strokeWeight(traceconf.traceweight);
          stroke(traceconf.tracecolor);
          line(i,min(max(yplot*yscale+yzero,ymin),ymax),i+1,min(max(yplot2*yscale+yzero,ymin),ymax));
        }
        
        //This was replaced by the above for purposes of resolution. 
        /*float x =  minXtick+(i-xmin)/xscale;
        float y = -trace.computePoint(x);
        float x2 =  minXtick+(i+1-xmin)/xscale;
        float y2 = -trace.computePoint(x2);
        // masking plot to pre-defined area
        //if((y*yscale+yzero>=ymin)&&(y*yscale+yzero<=ymax)&&(y2*yscale+yzero>=ymin)&&(y2*yscale+yzero<=ymax))
        if(((y*yscale+yzero>=ymin)&&(y*yscale+yzero<=ymax))||((y2*yscale+yzero>=ymin)&&(y2*yscale+yzero<=ymax)))
        {
          TraceConfig traceconfig = (TraceConfig)traceconfigArray.get(j);
          strokeWeight(traceconf.traceweight);
          stroke(traceconf.tracecolor);
          //line(i,y*yscale+yzero,i+1,y2*yscale+yzero);
          line(i,min(max(y*yscale+yzero,ymin),ymax),i+1,min(max(y2*yscale+yzero,ymin),ymax));
        }*/
      }
    }
  }
}
